/*      CaS, performs conversion from textfiles to .CAS files
 *
 * (c) Copyright 1996 Tom Wheeley <tomw@tsys.demon.co.uk>  All Rights Reserved
 *
 * You may freely use parts of this source code, provided that you:
 *  + Acknowledge the fact you used CaS source code in your program, visibly
 *    to your program's users.
 *  + Send the author of CaS a copy of your program.
 *  + You are not under obligation to include source for your program, but
 *    that is preferred.
 *
 */
 
#include "cas.h"
#include "misc.h"
#include "args.h"
#include "token.h"
#include "list.h"
#include "ftxt.h"
#include "fcas.h"
#include "ffxp.h"
#include "fbmp.h"
#include "fgif.h"
#include "fcom.h"
#include "ssconv.h"
#include "main.h"

unsigned char pow2[8]={1,2,4,8,0x10,0x20,0x40,0x80};
flagtype flags={
  "",
  {FALSE,fUnknown,""},            /* list {set,format,num,tokfile} */
  {fUnknown,NULL,"",SUBF_UNKNOWN},  /* output {format,filename,tokfile} */
  {fUnknown,NULL,"",SUBF_UNKNOWN},  /* input  {format,filename,tokfile} */
  MDL_UNKNOWN,                      /* model */
  0,                                /* conversions */
  FALSE,                            /* verbose */
  FALSE,                            /* pager */
  NULL                              /* debug */
};

#define CAS_BANNER  "This is CaS, Version 2.7183 (" CAS_VER_STR ")"
/* due apologies to TeX, Version 3.1415 (C Version 6.1) :-) */

int main(int argc, char *argv[])
{
  char *argw=NULL;           /* writeable copy of *argv */
  char *fmt=NULL;            /* temporary string for format specifier */
  tokenlist *toklis=NULL;    /* List of tokens */
  casinfo ci;           /* Abstract form of .cas header */
   
/* Process command line */

  flags.argv0=*argv;

  if (argc==1) {
    printf("%s\n",CAS_BANNER);
    exit(1);
  }

  for (argv++,argc--;argc;argv++,argc--) {
    argw=*argv;
    if ((*argw=='-') && argw[1]) {
      switch (argw[1]) {

/* Display help info */

        case '?': case 'h': 
          printf("%s\n",CAS_BANNER);
          printf("usage: CaS [-i[=type]] infile [-o[=type] outfile] [-l[=type]] [options]\n\n");
          printf("infile     file to process\n");
          printf("outfile    output file to create\n");
          printf("-l[=type]  list file to stdout\n");
          printf("-i[=type]  force type and specify input file to type\n");
          printf("-o[=type]  select type of and specify output file\n");
          printf("types\n");
          printf("  asc      plain ascii (for list, same as -l)\n");
          printf("  dos      dos codepage 437 text\n");
          printf("  cas      casio program (.cas)\n");
          printf("  fxp      fxp program (.fxp)\n");
          printf("  bmp      Windows bitmap (.bmp)\n");
          printf("  gif      Graphics Interchange Format (.gif)\n");
          printf("  com      serial communications to calculator\n");
          printf("options\n");
          printf("-v         verbose mode\n");
          printf("-p         pager\n");
          printf("-d[=file]  display debug information [in file]\n");
          printf("\nSee \"http://www.york.ac.uk/~tw104/\" for more info and updates.\n");
          printf("Made with pride in Great Britain.\n");

          exit(EXIT_CMD);

/* Set List output format */

        case 'l':   /* specifies list type */
          if (flags.list.set) Msg(EXIT_CMD,"main","Duplicate option: %s",argw);
          flags.list.set=TRUE;

          flags.list.argv=ArgsSplit(argw+2,&fmt);
          ArgsGetFormat(fmt,&flags.list.format,&flags.list.tokfile);
          break;

/* Override input filetype format heuristic */

        case 'i':   /* specifies input filetype and filename */
          if (flags.input.filename) Msg(EXIT_CMD,"main","Duplicate option: %s",argw);

          flags.input.argv=ArgsSplit(argw+2,&fmt);
          ArgsGetFormat(fmt,&flags.input.format,&flags.input.tokfile);

          argc--,argv++;
          if (!argc) Msg (EXIT_CMD,"main","input file not specified");
          flags.input.filename=*argv;
          break;
          
/* Set output filename and filetype (-oX filename) */

        case 'o':	/* specifies output filetype and filename */
          if (flags.output.filename) Msg(EXIT_CMD,"main","Duplicate option: %s",argw);

          flags.output.argv=ArgsSplit(argw+2,&fmt);
          ArgsGetFormat(fmt,&flags.output.format,&flags.output.tokfile);

          argc--,argv++;
          if (!argc) Msg (EXIT_CMD,"main","output file not specified");
          flags.output.filename=*argv;
          break;
          
/* Set Model */

        case 'm': 
          if (flags.model!=MDL_UNKNOWN) Msg(EXIT_CMD,"main","Duplicate option: %s",argw);

          ArgsSplit(argw+2,&fmt);
          if (!strnicmp(fmt,"cfx",3)) fmt+=3;
          if (!strnicmp(fmt,"fx",2)) fmt+=2;

          if (!strcmp(fmt,"7700") || !strcmp(fmt,"7")) {
            flags.model=MDL_7700;
            break;
          }
          if (!strcmp(fmt,"9700") || !strcmp(fmt,"9")) {
            flags.model=MDL_9700;
            break;
          }
          if (!strcmp(fmt,"9800") || !strcmp(fmt,"8")) {
            flags.model=MDL_9800;
            break;
          }
          if (!strcmp(fmt,"9850") || !strcmp(fmt,"5")) {
            flags.model=MDL_9850;
            break;
          }
          if (!strcmp(fmt,"9950") || !strcmp(fmt,"5")) {
            flags.model=MDL_9950;
            break;
          }
          if (!stricmp(fmt,"any") || !strcmp(fmt,"*")) {
            flags.model=MDL_UNKNOWN;
            break;
          }

          Msg(EXIT_CMD,"main","Unknown model: `%s'",argw);
          break;

/* set miscellaneous options */

        case 'v': flags.verbose=TRUE;break;     /* Verbose mode */
        case 'd': flags.debug=stderr;
                  argw+=2;if (*argw=='='||*argw==':') argw++;
                  if (*argw) flags.debug=tsys_fopen(argw,"w",FALSE);
                  if (!(flags.debug))
                    Msg(EXIT_CMD,"main","unable to open debug file `%s'",argw);
                  Msg(M_DEBUG,"main","debug messages active from -d switch");
                  break;                        
        case 'p': flags.pager=TRUE;break;       /* Screen Pager (more) */
        case 'c': flags.conv=CONV_BEFORE;break;
        case 'C': flags.conv=CONV_AFTER;break;

        default : Msg(EXIT_CMD,"main","Unknown option: %s",*argv);
      }
    } else {

/* Set input filename */

      if (flags.input.filename&&*flags.input.filename) {
        Msg(EXIT_CMD,"main","Filename (%s) already specified.  Unknown option: %s",flags.input.filename,*argv);
      } else {
        flags.input.filename=*argv;
      }
    }
  }
  
  if (flags.verbose) printf("%s\n\n",CAS_BANNER);

/* Check validity of switches */

  if (!flags.input.filename) {
    Msg(EXIT_FILE,"main","No input file");
  }
  if (!flags.output.filename && !flags.list.set) {
    flags.verbose=TRUE;
  }
 
/* Heuristic to decide type of input file, if not over-ridden with -i */

  if (flags.input.format==fUnknown)
    GetFileType(flags.input.filename,&flags.input.format,&flags.input.tokfile);
  if (flags.input.format==fUnknown)
    Msg(EXIT_FILE ,"main","Undetermined input filetype");
    /* should this default to ascii ? */

/* Default for list is ascii */

   if (flags.list.set && flags.list.format==fUnknown) {
     flags.list.format=ftxt;
     flags.list.tokfile=TOKEN_TOKFILE_ASC;
   }

/* Heuristic to decide type of output file if unspecified */

  if (flags.output.filename!=NULL && flags.output.format==fUnknown) {
    if (!GetFileTypeByExt(flags.output.filename,&flags.output.format,&flags.output.tokfile)) {
      switch (flags.input.format) {
        case ftxt: 
        case fBMP:
        case fGIF: flags.output.format=fCAS;break;
        case fCAS:
        case fFXP:
        default:   flags.output.format=ftxt;
                   flags.output.tokfile=TOKEN_TOKFILE_ASC;
      }
      Msg(M_DEBUG,"main","output filetype: %s (deduced from input type: %s)",
        FileType(flags.output.format),FileType(flags.input.format));
    }
  }

  Msg(M_DEBUG,"main","infile : %s (%s)",flags.input.filename,
                                        FileType(flags.input.format));
  Msg(M_DEBUG,"main","outfile: %s (%s)",flags.output.filename,
                                        FileType(flags.output.format));
  Msg(M_DEBUG,"main","list: (%s)",FileType(flags.list.format));

/* Load tokens into memory */

  Msg(M_DEBUG,"token"," input tokfile: %s",flags.input.tokfile);
  Msg(M_DEBUG,"token","output tokfile: %s",flags.output.tokfile);
  Msg(M_DEBUG,"token","  list tokfile: %s",flags.list.tokfile);

  toklis=InitTokens();

/* Initialise memory */

  ci.type=tUnknown;
  ci.data=tsys_calloc(CAS_MEMSIZE);
  ci.dataend=(char *)ci.data+CAS_MEMSIZE-1;
  ci.length=0;
  ci.size1=ci.size2=0;
  ci.aux=NULL;

/* Read data from input file */

  switch (flags.input.format) { 
    case ftxt: TXTReadData(flags.input.filename,&ci,toklis);break;
    case fCAS: CASReadData(flags.input.filename,&ci);break;
    case fFXP: FXPReadData(flags.input.filename,&ci);break;
    case fBMP: BMPReadData(flags.input.filename,&ci);break;
    /* GIF reading not yet supported */
    case fCOM: COMReadData(flags.input.filename,&ci);break;
    default  : Msg(EXIT_BUG,"main","Input format not yet supported");
  } 

/* Perform conversions (before) */

  if (flags.conv==CONV_BEFORE)
    switch (ci.type) {
      case tSSMono: SSConv(&ci);break;

      default:
        Msg(M_WARN,"convert","Cannot convert this file type: %s",ListType(ci.type));
    }

/* List info */

  Msg(M_VERBOSE,"File Type","%s",ListType(ci.type));

  if (flags.list.set) {
    if (ArgsIsSet("dump",flags.list.argv)) {
      ListDump(&ci,toklis);
      if (ArgsIsSet("list",flags.list.argv)) {
        ListData(&ci,toklis);
      }
    } else {
      ListData(&ci,toklis);
    }
  }

/* Perform conversions (after) */

  if (flags.conv==CONV_AFTER)
    switch (ci.type) {
      case tSSMono: SSConv(&ci);break;

      default:
        Msg(M_WARN,"convert","Cannot convert this file type: %s",ListType(ci.type));
    }

/* Write data to output file */

  if (flags.output.filename) {
    switch (flags.output.format) {
      case ftxt: TXTWriteData(flags.output.filename,&ci,toklis);break;
      case fCAS: CASWriteData(flags.output.filename,&ci);break;
      case fFXP: FXPWriteData(flags.output.filename,&ci);break;
      case fBMP: BMPWriteData(flags.output.filename,&ci);break;
      case fGIF: GIFWriteData(flags.output.filename,&ci);break;
      case fCOM: COMWriteData(flags.output.filename,&ci);break;
      default  : Msg(EXIT_BUG,"main","Output format not yet supported");
    }
  }

/* Tidy up */

  if (flags.input.format==ftxt) TXTReadToken(NULL,NULL,NULL,-1);
          /* displays error if final char was bad */

  FreeTokens(toklis);

  tsys_fclose(flags.debug);

  return EXIT_SUCCESS; 
}


/*
 * GetFileType.  Guess the file type of an input file.  Needs improvement.
 */

bool GetFileType(char *filename, formats *format_r, char **tokfile_r)
{
  FILE *f;
/* casheader ch; */
  char colon='!';

  if (GetFileTypeByExt(filename, format_r, tokfile_r)) return TRUE;

  /* no/unrecognised extension.  Look at contents */

  if (!(f=fopen(filename,"rb"))) {
    Msg(M_DEBUG,"filetype","unable to open file");
    return FALSE;
  }
 
  colon=getc(f);
 
  if (colon == ':') {
    Msg(M_DEBUG,"filetype","CAS file (':')");
    *format_r=fCAS;
    *tokfile_r="";
    fclose(f);
    return TRUE; 
  }

  if (colon=='B')
    if (getc(f)=='M') {
    Msg(M_DEBUG,"filetype","BMP file ('BM')");
    *format_r=fBMP;
    *tokfile_r="";
    fclose(f);
    return TRUE; 
  }

  Msg(M_DEBUG,"filetype","Text file ('%c')",colon);
  *format_r=ftxt;
  *tokfile_r=TOKEN_TOKFILE_ASC;
  fclose(f);
  return TRUE;
}


bool GetFileTypeByExt(char *filename, formats *format_r, char **tokfile_r)
{
  char *extn=NULL;

  Msg(M_DEBUG,"filetype/ext","Checking `%s'",filename);

  /* check for serial devices */

#ifdef CAS_DOS
  if (!strnicmp(filename,"COM",3) && !filename[4] &&
                                 (filename[3]>'0' && filename[3]<'5')) {
    Msg(M_DEBUG,"filetype/ext","COM port (COMx)");
    *format_r=fCOM;
    *tokfile_r="";
    return TRUE;
  }
#endif /* CAS_DOS */

#ifdef CAS_UNIX
  if (!strnicmp(filename,"/dev/tty",8)) {
    Msg(M_DEBUG,"filetype/ext","Serial device (/dev/ttyX)");
    *format_r=fCOM;
    *tokfile_r="";
    return TRUE;
  }
#endif /* CAS_UNIX */

  /* check stdin, stdout, stderr */
  
  if (!strcmp(filename,"-") ||
      !strcmp(filename,"stdin") ||
      !strcmp(filename,"stdout") ||
      !strcmp(filename,"stderr")) {
    Msg(M_DEBUG,"filetype/ext","standard file stream => ascii");
    *format_r=ftxt;
    *tokfile_r=TOKEN_TOKFILE_ASC;
    return TRUE;
  }
  
  /* check extensions */
  
  extn=strchr(filename,'.');
  if (extn && *extn) {
    extn++;
    if (!stricmp(extn,"cas")) {
      Msg(M_DEBUG,"filetype/ext","CAS file (.cas)");
      *format_r=fCAS;
      *tokfile_r="";
      return TRUE;
    }
    if (!stricmp(extn,"txt")) {
      Msg(M_DEBUG,"filetype/ext","TXT file (.txt)");
      *format_r=ftxt;
      *tokfile_r=TOKEN_TOKFILE_ASC;
      return TRUE;
    }
    if (!stricmp(extn,"fxp")) {
      Msg(M_DEBUG,"filetype/ext","FXP file (.fxp)");
      *format_r=fFXP;
      *tokfile_r="";
      return TRUE;
    }
    if (!stricmp(extn,"bmp")) {
      Msg(M_DEBUG,"filetype/ext","BMP file (.bmp)");
      *format_r=fBMP;
      *tokfile_r="";
      return TRUE;
    }
    if (!stricmp(extn,"gif")) {
      Msg(M_DEBUG,"filetype/ext","GIF file (.gif)");
      *format_r=fGIF;
      *tokfile_r="";
      return TRUE;
    }
  }

  Msg(M_DEBUG,"filetype/ext","Unrecognised");

  return FALSE;
}


char *FileType(formats f)
{
   switch (f) {
     case ftxt: return "Text";
     case fCAS: return "CAS";
     case fFXP: return "FXP";
     case fBMP: return "Bitmap";
     case fGIF: return "GIF";
     case fCOM: return "Serial Comms";
     default:   return "Unknown";
   }
}

 /* end of main.c */
