/*      CaS, performs conversion from textfiles to .CAS files
 *
 * (c) 1996 Tom Wheeley <tomw@tsys.demon.co.uk>  All Rights Reserved
 *
 * You may freely use parts of this source code, provided that you:
 *  + Acknowledge the fact you used CaS source code in your program, visibly
 *    to your program's users.
 *  + Send the author of CaS a copy of your program.
 *  + You are not under obligation to include source for your program, but
 *    that is preferred.
 *
 */


#include "cas.h"
#include "misc.h"

extern flagtype flags;

void Msg (int type, const char *section, const char *msgstr, ...)
{
  FILE *outfile=stdout;
  va_list args;
  
  if ((type==M_DEBUG) && !flags.debug) return;
  if ((type==M_VERBOSE) && !flags.verbose) return;
  
  Pager(PAGER_ADD);
  
  switch (type) {
    case M_MSG:
    case M_VERBOSE: break;
    case M_WARN:  outfile=stderr; fprintf(outfile,"warning: ");break;
    case M_ERROR: outfile=stderr; fprintf(outfile,"error: ");break;
    case M_DEBUG: outfile=flags.debug; fprintf(outfile,"debug: ");break;
    default:      outfile=stderr; fprintf(outfile,"fatal:%d: ",type);
  }
  fprintf(outfile,"%s: ",section);

  va_start (args,msgstr);
  vfprintf(outfile,msgstr,args);
  va_end (args);

  putc('\n',outfile);
  
  if (type>0) exit(type);
}


void Pager(int dowhat)
{
  static int count=0;
  char buf[99];
  
/*  return; */
  
  if (dowhat==PAGER_CLEAR) {
    count=0;
    return;
  }

  if (!flags.pager && (!(dowhat&PAGER_FORCE)))
    return;
  
  if (!(dowhat&PAGER_NOADD)) ++count;

  if (count>=PAGER_LINES||(dowhat&PAGER_NOW)) {
    if (!(dowhat&PAGER_NORESET)) count=0;
    if (!(dowhat&PAGER_NOMSG))
      fprintf(stderr,"pager: press <enter> to continue. ");
    if (!(dowhat&PAGER_NOPAUSE)) fgets(buf,99,stdin);
  }
}

/*
 *  Memory allocation routines with error checking, reporting and exit
 */

void *tsys_malloc (size_t size)
{
  void *memptr=NULL;
  
  if ((memptr=malloc(size)) == NULL)
      Msg(EXIT_MEM,"malloc","unable to allocate %ld bytes",size);
  return(memptr);
}


void *tsys_calloc (size_t size)
{
  void *memptr=NULL;
  
  if ((memptr=calloc(1,size)) == NULL)
      Msg(EXIT_MEM,"calloc","unable to allocate %ld bytes",size);
  return(memptr);
}


void *tsys_realloc (void *oldptr, size_t size)
{
  void *memptr=NULL;
  
  if ((memptr=realloc(oldptr, size)) == NULL)
      Msg(EXIT_MEM,"realloc","unable to allocate %ld bytes",size);
  return(memptr);
}



FILE *tsys_fopen(char *fname, char *faccess, int owndir)
{
  char newname[2 * FILENAME_MAX+1];
  char *eod=NULL;
  FILE *handle=NULL;

  if (!fname||!faccess) return NULL;

  if (!strcmp(fname,"stdin")) return stdin;
  if (!strcmp(fname,"stdout")) return stdout;
  if (!strcmp(fname,"stderr")) return stderr;
  if (!strcmp(fname,"-")) {
    switch(*faccess) {
      case 'w': return stdout;
      case 'r': return stdin;
      default: return NULL;
    }
  }

  if ((handle=fopen(fname,faccess))) return handle;
  if (!owndir) return NULL;

  strcpy(newname,flags.argv0);
  if (!(eod=strrchr(newname,CAS_DIRSEP)))
    return NULL;    
  strcpy(eod+1,fname);

  Msg(M_DEBUG,"fopen","attempting to open `%s'",newname);
  
  if ((handle=fopen(newname,faccess))) return handle;

  return NULL;
}


int tsys_fclose(FILE *handle)
{
  if (!handle) return EOF;
  if (handle==stdin) return 0;
  if (handle==stdout) return 0;
  if (handle==stderr) return 0;

  return fclose(handle);
}


/*
 * Calculate checksums
 */

unsigned char CalcCSum(void *data, int size)
{
  unsigned char cs=0;

  if (!data || size<=0) return 0;

  for (;size;)
    cs+=256-((unsigned char*)data)[--size];

  cs &= 0xff;

  Msg(M_DEBUG,"csum","calculated checksum: 0x%02x",cs);
  return cs;
}

/*
 *  Wildcard matching function.  Only works with * and ? so far :(
 */

int WildMatch(char *wildcard, char *teststr)
{
  char *nextw=NULL;
  char *nextt=NULL;
  
  if (!*teststr && !*wildcard) return TRUE;
  if (!*teststr || !*wildcard) return FALSE;
  
  if (*wildcard=='?'||(*teststr==*wildcard)) {
    return(WildMatch(++wildcard,++teststr));
  }
  if (*wildcard=='*') {
    nextw=wildcard;
    if (!*++nextw) return TRUE;
    if (!(nextt=strchr(teststr,*nextw))) return FALSE;
    if (WildMatch(nextw,nextt)) return TRUE;
    return (WildMatch(wildcard,++nextt));
  }
  return FALSE;
}



/* Adapted from K&R2, page 144.  My formula is random... */

unsigned long Hash(char *s, unsigned max)
{
 unsigned long hashval;

 if (!s||max==0) return 0;
 for (hashval=0;*s;s++)
   hashval = *s * 7 + hashval * 15;
 return (hashval%max);
}


/*
 * Convert numbers depending on endianness.  Stolen from QEU 0.3.
 */
#ifdef FAT_ENDIAN

ulongint SwapInt32(ulongint x)
{
  return (((x << 24) & 0xff000000)
          | ((x<< 8) & 0x00ff0000)
          | ((x>> 8) & 0x0000ff00)
          | ((x>>24) & 0x000000ff));
}

#endif /* FAT_ENDIAN */

/* end of misc.c */

