/*      CaS, performs conversion from textfiles to .CAS files
 *
 * (c) 1996 Tom Wheeley <tomw@tsys.demon.co.uk>  All Rights Reserved
 *
 * You may freely use parts of this source code, provided that you:
 *  + Acknowledge the fact you used CaS source code in your program, visibly
 *    to your program's users.
 *  + Send the author of CaS a copy of your program.
 *  + You are not under obligation to include source for your program, but
 *    that is preferred.
 *
 */

/*
 * SSCONV.C  Screenshot to Program conversion routine.
 *           Originally written by Tom Lynn.  Ported to C by Tom Wheeley
 */

#include "cas.h"
#include "misc.h"
#include "args.h"
#include "token.h"
#include "list.h"
#include "ssconv.h"

unsigned char *SSConv(casinfo *info)
{
  int spx,spy,ipx,ipy,npx,npy,cpx,cpy;
  int x,y,spoint;
  bool foundend;
  bool oddstep;
  unsigned char (*image)[SS_XMAX][SS_YMAX];
  unsigned char (*notdone)[SS_XMAX][SS_YMAX];
  unsigned char (*pretend)[SS_XMAX][SS_YMAX];
  unsigned char *program;
  prginfo *pi;
  short sqx[]={ 0, 2, 0,-2,-2, 2, 2,-2,-1, 1, 2, 2, 1,-1,-2,-2};
  short sqy[]={-2, 0, 2, 0,-2,-2, 2, 2,-2,-2,-1, 1, 2, 2, 1,-1};
  short  ix[]={ 0, 1, 0,-1,-1, 1, 1,-1,-1, 0, 1, 1, 0,-1,-1,-1};
  short  iy[]={-1, 0, 1, 0,-1,-1, 1, 1,-1,-1, 0, 0, 1, 1, 1,-1};

  if (info->type!=tSSMono) {
    Msg(EXIT_BUG,"ssconv","Unsupported file variation: %s",
           ListType(info->type));
    return NULL;
  }

  /* set up memory regions */

  image=info->data;
  notdone=tsys_malloc(SS_XMAX * SS_YMAX);
  pretend=tsys_malloc(SS_XMAX * SS_YMAX);
  info->data=program=tsys_malloc(CAS_MEMSIZE);
  info->aux=pi=tsys_malloc(sizeof(prginfo));

  memcpy(notdone,image,SS_XMAX * SS_YMAX);
  memcpy(pretend,image,SS_XMAX * SS_YMAX);

  SSRange(info->size1-1,info->size2-1,&program);

  /* Find first black point */

  /* We only go to width-1 and height-1 as we can only draw on a slightly
   * smaller screen (1 pixel lines at top and left unavailable)
   */

  x=info->size1;
  for (y=info->size2-2;y>=0;y--)     /* info->size2 = height */
    for (x=1;x<info->size1;x++)         /* info->size1 = width */
      if ((*notdone)[x][y]) goto gotpoint;
  gotpoint:

  while(x<info->size1 && y<info->size2-1) {
    foundend=FALSE;
    do {
      spoint=0;
      cpx=x;cpy=y;
      do {
        spx=x+sqx[spoint];
        spy=y+sqy[spoint];
        if (spx<info->size1 && spx>0 && spy<info->size2-1 && spy>=0 &&
                    ( (foundend && (*notdone)[spx][spy])
                   || (!foundend && (*pretend)[spx][spy]) )  ) {

          ipx=x+ix[spoint];    /* check intermediate points */
          ipy=y+iy[spoint];
          if ((*image)[ipx][ipy]) {
            /* Draw a line! */
            if (foundend) {
              SSPlot(x,y+1,&program);
              (*notdone)[x][y]=0;
              (*notdone)[ipx][ipy]=0;
            }
            (*pretend)[x][y]=0;
            (*pretend)[ipx][ipy]=0;

            npx=spx;npy=spy;
            oddstep=TRUE;
            do {
              cpx=npx;cpy=npy;
              if (foundend) {
                (*notdone)[cpx][cpy]=0;
              }
              (*pretend)[cpx][cpy]=0;
              if (oddstep) {
                npx=cpx+ix[spoint];
                npy=cpy+iy[spoint];
              } else {
                npx=cpx+sqx[spoint]-ix[spoint];
                npy=cpy+sqy[spoint]-iy[spoint];
              }
              oddstep=!oddstep;
            } while ((npx>0 && npx<info->size1 && npy>=0 && npy<info->size2-1)
                    && (*image)[npx][npy]);
            if (foundend) {
              SSPlot(cpx,cpy+1,&program);
              SSLine(&program);
              (*notdone)[cpx][cpy]=0;
            }
            (*pretend)[cpx][cpy]=0;
            if (npx>0 && npx<info->size1 && npy>=0 && npy<info->size2-1) {
              spoint=0;
              x=cpx;y=cpy;
            } else {
              spoint=16;
              x=cpx;y=cpy;
            }
          } else {
            ++spoint;
          }
        } else {
          ++spoint;
        }
      } while (spoint<8);
      foundend=!foundend;
      x=cpx;y=cpy;
    } while (foundend);

    if ((*notdone)[x][y]) {
      SSPlot(x,y+1,&program);
      (*notdone)[x][y]=0;
      (*pretend)[x][y]=0;
    }

  for (y=info->size2-2;y>=0;y--)     /* info->size2 = height */
    for (x=1;x<info->size1;x++)          /* info->size1 = width */
      if ((*notdone)[x][y]) goto gotnext;
  gotnext:
    ;
  }

  /* Clean up memory */

  free(notdone);
  free(pretend);
  free(image);

  /* Complete conversion to program type for CaS */

  SSFinish(&program);

  pi->progtype=0;
  pi->data=info->data;
  pi->size=(int)(program - pi->data);

  Msg(M_DEBUG,"ssconv","fin: %p, start: %p (%p), diff: %ld (%d)",
               program,pi->data,info->data,program-pi->data,pi->size);

  info->type=tSingleProg;
  info->size1=1;
  info->size2=pi->size+2;

  return pi->data;
}


void SSPlot(int x, int y, unsigned char **data)
{
  static int prevx=-1;
  static int prevy=-1;

  /* We can save memory by using the previously plotted position instead
     of the absolute position when possible */

  if (x==prevx && y==prevy) return;

  *((*data)++)=SS_TOK_PLOT;
  if (x==prevx && x>9) *((*data)++)='X'; else
    if (x==prevy && x>9) *((*data)++)='Y'; else 
       SSNum(x,data);
  *((*data)++)=SS_TOK_COMMA;
  if (y==prevy && y>9) *((*data)++)='Y'; else
    if (y==prevx && y>9) *((*data)++)='X'; else
      SSNum(y,data);
  *((*data)++)=SS_TOK_NL;

  prevx=x; prevy=y;
}


void SSLine(unsigned char **data)
{
  *((*data)++)=SS_TOK_LINE;
  *((*data)++)=SS_TOK_NL;
}


void SSRange(int width,int height, unsigned char **data)
{
  *((*data)++)=SS_TOK_RANGE;  *((*data)++)='1';
  *((*data)++)=SS_TOK_COMMA;  SSNum(width,data);
  *((*data)++)=SS_TOK_COMMA;  *((*data)++)='0';
  *((*data)++)=SS_TOK_COMMA;  *((*data)++)='1';
  *((*data)++)=SS_TOK_COMMA;  SSNum(height,data);
  *((*data)++)=SS_TOK_COMMA;  *((*data)++)='0';
  *((*data)++)=SS_TOK_NL;
}


void SSFinish(unsigned char **data)
{
  *((*data)++)=SS_TOK_PLOT;   *((*data)++)='0';
  *((*data)++)=SS_TOK_COMMA;  *((*data)++)='0';
  /*  *((*data)++)=SS_TOK_DISP;   Not req. at end of prog. */
  *((*data)++)=0xff;
}


void SSNum(int num, unsigned char **data)
{
  char s[20];
  char *t;

  sprintf(s,"%d",num);
  for (t=s;*t;t++)
    *((*data)++)=*t;

}

/* end of ssconv.h */
