/*      CaS, performs conversion from textfiles to .CAS files
 *
 * (c) 1996 Tom Wheeley <tomw@tsys.demon.co.uk>  All Rights Reserved
 *
 * You may freely use parts of this source code, provided that you:
 *  + Acknowledge the fact you used CaS source code in your program, visibly
 *    to your program's users.
 *  + Send the author of CaS a copy of your program.
 *  + You are not under obligation to include source for your program, but
 *    that is preferred.
 *
 */

/*
 * define CAS_TOKCHECK for debugging info. (as in tokcheck)
 */

#include "cas.h"
#include "misc.h"
#include "token.h"

extern flagtype flags;
static int clash=0;

/*
 * InitTokens.  Allocates memory for the arrays of pointers to the tokens
 *              and calls LoadTokens appropriately            
 */

tokenlist *InitTokens(void)
{
  int i;
  tokenlist *toklis=NULL;
  
  toklis=tsys_calloc(sizeof (tokenlist));
  Msg(M_DEBUG,"token","Allocated 0x%04lx bytes for table", sizeof (tokenlist)); 
  toklis->tokpos=toklis->tokblock=tsys_malloc(TOKEN_BLOCKSIZE);
  Msg(M_DEBUG,"token","Allocated 0x%04lx bytes for tokens", TOKEN_BLOCKSIZE); 
  toklis->tokend=toklis->tokblock + TOKEN_BLOCKSIZE;

  if (flags.input.format==ftxt) {
    LoadTokens(flags.input.tokfile,toklis,1,1,1);
  }

  if (flags.output.format==ftxt &&
      strcmp(flags.input.tokfile,flags.output.tokfile)) {
    for (i=0;i<TOKEN_MAX;toklis->out[i++]=NULL)
      ;
    if (!strcmp(flags.output.tokfile,flags.list.tokfile)) {
      for (i=0;i<TOKEN_MAX;toklis->list[i++]=NULL)
        ;
      LoadTokens(flags.output.tokfile,toklis,0,1,1);
    } else {
      LoadTokens(flags.output.tokfile,toklis,0,1,0);
    }
  }

  if (flags.list.set && flags.list.format==ftxt &&
      strcmp(flags.list.tokfile,flags.input.tokfile)) {
    for (i=0;i<TOKEN_MAX;toklis->list[i++]=NULL)
      ;
    LoadTokens(flags.list.tokfile,toklis,0,0,1);
  }

  return toklis;
}

/*
 * LoadTokens.  Loads all the tokens from a tokenfile into the respective
 *              structures determined by in, out and list.
 */

void LoadTokens(char *tokfile,tokenlist *toklis,int in,int out,int list)
{
  FILE *tf=NULL;
  char tokline[TOKEN_TOKSIZE];
  int count=0;
  int numtok=0;
  int linecount;

  if (!(tf=tsys_fopen(tokfile,"r",TRUE))) {
    Msg(EXIT_FILE,"token","unable to read token file `%s'",tokfile);
  }
  
  for(;;) {
    if (feof(tf)) break;
    fgets(tokline,TOKEN_TOKSIZE,tf);
    if (tokline[strlen(tokline)-1]=='\n')
      tokline[strlen(tokline)-1]='\0';
#ifdef CAS_TOKCHECK
    if (isspace(tokline[strlen(tokline)-1]))
      Msg(M_DEBUG,"token","token line ends in whitespace. `%s'",tokline);
#endif
    
    if (!*tokline || *tokline=='#') continue;
    
    if ((linecount=AddTokenLine(tokline,toklis,in,out,list))) numtok++;
    count+=linecount;
  }
  
  fclose(tf);
  
#ifndef CAS_TOKCHECK
  if (in) {
#endif
    Msg(M_DEBUG,"token","%s: read %d for %d; %d clashes. (i%s%s)",
                 tokfile,count,numtok,clash,out?"o":"",list?"l":"");
#ifndef CAS_TOKCHECK
  } else {
    Msg(M_DEBUG,"token","%s: read %d synonyms for %d tokens. (%s%s)",
                 tokfile,count,numtok,out?"o":"",list?"l":"");
  }
#endif
}


/*
 * AddTokenLine.  Adds a line of tokens to the internal structures, which
 *                dependingon the in,out,list parameters
 */

int AddTokenLine(char *tokline,tokenlist *toklis,
                 int in,int out,int list)
{
  long toknum=0;
  int count=0;
  int tflag;
  char *token=NULL;
  char *duptoken=NULL;

  if (!(token=strtok(tokline,"|"))) {
#ifdef CAS_TOKCHECK
    Msg(M_WARN,"token","invalid token line: %s",tokline);
#endif
    return 0;
  }

  toknum=strtol(token,NULL,0);          /* should this be base 10 ? */
  if (toknum<0 || toknum > TOKEN_MAX) {
#ifdef CAS_TOKCHECK
    Msg(M_WARN,"token","invalid code: %ld from `%s'",toknum,token);
#endif
    return 0;
  }
  
  while ((token=strtok(NULL,"|"))) {
    tflag=TokenFlag(&token);
    if (!(tflag&flags.model)) continue;
    duptoken=DupToken(token,toklis);
    if (in && !(tflag&TF_LISTONLY))
      AddToHash(duptoken,toknum|(tflag&FLAG_MASK),toklis);
    if (out && !toklis->out[toknum] && !(tflag&TF_LISTONLY))
      toklis->out[toknum]=duptoken;
    if (list)
      if (!toklis->list[toknum])
        toklis->list[toknum]=duptoken;
      else
        if (tflag&TF_LISTONLY)
          toklis->list[toknum]=duptoken;
    count++;
  }

  return count;
}


int TokenFlag(char **token_r)
{
  char *tf=NULL;
  char *temp=NULL;
  int tflag;

  if (**token_r!='{') return TF_DEFAULT;

  tf=(*token_r) + 1;
  for(;**token_r;(*token_r)++)
    if (**token_r=='}') {
      *((*token_r)++)='\0';
      break;
    }

  tflag=TF_DEFAULT;
  for (temp=tf;*temp;temp++)
    if (isdigit(*temp)) tflag=0;

  for (temp=tf;*temp;temp++)
    switch(tolower(*temp)) {
      case '7':tflag|=MDL_7700;break;
      case '9':tflag|=MDL_9700;break;
      case '8':tflag|=MDL_9800;break;
      case '5':tflag|=MDL_9850;break;
      case 'l':tflag|=TF_LISTONLY;break;
      case 'v':tflag|=TF_VERBATIM;break;
      case 'n':tflag|=TF_NNE;break;
      default:
#ifdef CAS_TOKCHECK
        Msg(M_DEBUG,"token","Invalid token flag: `%c'",*temp)
#endif
        ;
    }

  return tflag;
}


/*
 * Duptoken.  Duplicates a token, expanding control characters (@ prefixed)
 *            as it goes.  Returns pointer to new string
 */

char *DupToken(char *token,tokenlist *toklis)
{
  char *i=NULL;
  char *j=NULL;
  char *duptoken=NULL;
  long num;
  char ntmp[4];
  char base='d';

  duptoken=toklis->tokpos;
  if (toklis->tokpos + strlen(token) > toklis->tokend) {
    Msg(EXIT_BUG,"token","insufficient memory for tokens");
  }

  for (i=token,j=duptoken;*i;i++) {
    if (!isprint((unsigned char)*i)) { 
      /* Allow `extended ascii characters' -- to make CP437.tok readable */
#ifdef CAS_DOS
      if ((unsigned char) (*i) < 128) {
#else
      if (TRUE) {
#endif
#ifdef CAS_TOKCHECK
        if (*i!='\r') Msg(M_DEBUG,"token","Skipping unprintable: %u",(unsigned char) *i);
				/* We don't want a warning under Unix if */
				/* there's a CR at the end of every line! */
#endif
        continue;
      }
    }
    if (*i=='@') {
      if (*++i=='\\') {
        switch (*++i) {
          case 'n': *(j++)='\n';break; /* newline */
          case 't': *(j++)='\t';break; /* tab */
          case '@': *(j++)='@';break;  /* Token file control chars */
          case '$': *(j++)='$';break;  
          case '{': *(j++)='{';break;
          case 'p': *(j++)='|';break;
          case '0': *(j++)='\0';break; /* generally for blank tokens */
                                       /* could also be used for comments? */
          default:
#ifdef CAS_TOKCHECK
            Msg(M_DEBUG,"token","incorrect character code `@\\%c' in %s",*i,token)
#endif
          ;
        }
      } else {
        if (*i=='h'||*i=='x'||*i=='o'||*i=='d') base=*(i++);
        strncpy(ntmp,i,3);i+=strlen(ntmp);
#ifdef CAS_TOKCHECK
        if (strlen(ntmp)!=3 || !isdigit(ntmp[1])) {
          Msg(M_DEBUG,"token","incorrectly formatted @ value `%s' (%d) in `%s'"
                            ,ntmp,strlen(ntmp),token);
        }
#endif
        num=strtol(ntmp,NULL,(base=='o')?8:(base=='d')?10:16);
#ifdef CAS_TOKCHECK
        if (num>255||num<1)
            Msg(M_DEBUG,"token","@ value out of range (%ld) from %s in %s",num,ntmp,token);
#endif
        *(j++)=(char) num;
        i--;
      }
    } else {  /* Normal, everyday character... */
      *(j++)=*i;
    }
  }
  *j='\0';
  toklis->tokpos=j+1;

  return duptoken;
}

/*
 * AddToHash.  Adds token to Hash table (toklis->in).  Returns hash value
 */

unsigned long AddToHash(char *token,int toknum,tokenlist *toklis)
{
  unsigned long hashval;

  hashval=Hash(token,TOKEN_HASHSIZE);
  while (toklis->in[hashval].code) {
    if (!strcmp(token,toklis->in[hashval].token)) {
      if ((toklis->in[hashval].code&TOKEN_MASK)==(toknum&TOKEN_MASK)) {
#ifdef CAS_TOKCHECK
        Msg(M_DEBUG,"token","duplicate token `%s' for same code %d.",token,toknum&TOKEN_MASK);
#endif
        return 0;
      } else {
#ifdef CAS_TOKCHECK
          Msg(M_DEBUG,"token",
            "duplicate token `%s', previous code %d, ignoring new code %d.",
            token,toklis->in[hashval].code&TOKEN_MASK,toknum&TOKEN_MASK);
#endif
        return 0;
      }
    }
#ifdef CAS_TOKCHECK
      Msg(M_DEBUG,"token","hash clash with `%s' adding `%s'",
                            toklis->in[hashval].token,token);
#endif
    ++clash;
    hashval++;
    hashval %= TOKEN_HASHSIZE;
  }

  toklis->in[hashval].code=toknum;
  toklis->in[hashval].token=token;

  return hashval;
}


/*
 * FreeTokens.  Frees memory used by token lists
 */

void FreeTokens(tokenlist *toklis)
{
  if (!toklis) return;

  free(toklis->tokblock);
  free(toklis);
}


/*
 * CodeToToken.  Returns the token for a ctoken.  If display is set returns
 *               display token (list),  else returns program token (out)
 */

char *CodeToToken(int ctoken, tokenlist *toklis, bool display)
{
  static char unknown[]="@\0\0\0\0\0";
  char *num;
  
  if (display && toklis->list[ctoken])
    return toklis->list[ctoken];
  if (toklis->out[ctoken])
    return toklis->out[ctoken];

  num=&unknown[1];
  sprintf(num,"d%03u",ctoken);
  return unknown;
}


/*
 * TokenToCode.  Returns the ctoken for a string using the hash table
 */

int TokenToCode(char *token,tokenlist *toklis,int verbatim)
{
  unsigned long hashval;
  long num;
  char *check=NULL;

/* @bnnn converts a number to an ascii value.  Optional b sets base. */

  if (strlen(token)>1 && token[0]=='@') {

    if (isdigit(token[1])) {
      if (strlen(token)==4) {
        num=strtol(token+1,&check,10);
        if (!*check) {
          if (num>TOKEN_MAX||num<0)
            Msg(M_ERROR,"parse","@ value out of range (%ld) in %s",num,token);
          return ((int) num);
        }
      }
    } else {
      switch (token[1]){
        case 'x': case 'h': case 'o': case 'd':
          if (strlen(token)==5) {
            num=strtol(token+2,&check,(token[1]=='o')?8:(token[1]=='d')?10:16);
            if (!*check) {
              if (num>TOKEN_MAX||num<=0)
                Msg(M_ERROR,"parse","@ value out of range (%ld) in %s",num,token);
              return ((int) num);
            }
          }
          /* case '@': return 255; obsolete? */
        default:;
      }
    }
  }

  hashval=Hash(token,TOKEN_HASHSIZE);
  while (toklis->in[hashval].token && strcmp(token,toklis->in[hashval].token)) {
    hashval++;
    hashval %= TOKEN_HASHSIZE;
  }

  if (verbatim && (toklis->in[hashval].code & TF_VERBATIM)) {
    Msg(M_DEBUG,"parse","rejecting non-verbatim `%s'",token);
    return 0;
  }

  return toklis->in[hashval].code & TOKEN_MASK;
}

/* end of token.c */
